from enfobench.evaluation.client import ForecastClient
from enfobench.evaluation.evaluate import (
    cross_validate,
    evaluate_metrics,
)
from enfobench.evaluation.model import AuthorInfo, ForecasterType, Model, ModelInfo
from enfobench.evaluation.protocols import Dataset

__all__ = [
    "ForecastClient",
    "cross_validate",
    "evaluate_metrics",
    "Dataset",
    "Model",
    "ModelInfo",
    "AuthorInfo",
    "ForecasterType",
]



import logging
from abc import ABCMeta
from dataclasses import dataclass
from pathlib import Path

import duckdb
import pandas as pd

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class Metadata:
    unique_id: str
    location_id: str
    latitude: float
    longitude: float
    building_type: str


class Subset(metaclass=ABCMeta):  # noqa: B024
    """Subset class representing one of the subset of the HuggingFace dataset.

    Args:
        file_path: The path to the subset file.
    """

    def __init__(self, file_path: Path | str) -> None:
        file_path = Path(file_path).resolve()
        if not file_path.is_file() or not file_path.exists():
            msg = "Please provide the existing file where the subset is located."
            raise ValueError(msg)
        self.file_path = file_path

    def __repr__(self):
        return f"{self.__class__.__name__}(file_path={self.file_path})"

    def read(self) -> pd.DataFrame:
        """Read the subset from the file."""
        return pd.read_parquet(self.file_path)


class MetadataSubset(Subset):
    """Metadata subset of the HuggingFace dataset containing all metadata about the meters.

    Args:
        file_path: The path to the subset file.
    """

    def list_unique_ids(self) -> list[str]:
        """Lists all unique ids."""
        query = """
            SELECT DISTINCT unique_id
            FROM read_parquet(?)
        """
        conn = duckdb.connect(":memory:")
        return conn.execute(query, parameters=[str(self.file_path)]).fetch_df().unique_id.tolist()

    def get_by_unique_id(self, unique_id: str) -> Metadata:
        """Returns the metadata for the given unique id.

        Args:
            unique_id: The unique id of the meter.
        """
        query = """
            SELECT *
            FROM read_parquet(?)
            WHERE unique_id = ?
        """
        conn = duckdb.connect(":memory:")
        df = conn.execute(query, parameters=[str(self.file_path), unique_id]).fetch_df()
        if df.empty:
            msg = f"Unique id '{unique_id}' was not found."
            raise KeyError(msg)
        return Metadata(**df.to_dict(orient="records")[0])


class WeatherSubset(Subset):
    """Weather subset of the HuggingFace dataset containing all weather data.

    Args:
        file_path: The path to the subset file.
    """

    def list_location_ids(self) -> list[str]:
        """Lists all location ids."""
        query = """
            SELECT DISTINCT location_id
            FROM read_parquet(?)
        """
        conn = duckdb.connect(":memory:")
        return conn.execute(query, parameters=[str(self.file_path)]).fetch_df().location_id.tolist()

    def get_by_location_id(self, location_id: str, columns: list[str] | None = None) -> pd.DataFrame:
        """Returns the weather data for the given location id.

        Args:
            location_id: The location id of the weather station.
            columns: The columns to return. If None, all columns are returned.
        """
        conn = duckdb.connect(":memory:")

        if columns:
            query = f"""
                SELECT timestamp, {", ".join(columns)}
                FROM read_parquet(?)
                WHERE location_id = ?
            """  # noqa: S608
        else:
            query = """
                SELECT *
                FROM read_parquet(?)
                WHERE location_id = ?
            """
        df = conn.execute(query, parameters=[str(self.file_path), location_id]).fetch_df()
        if df.empty:
            msg = f"Location id '{location_id}' was not found."
            raise KeyError(msg)

        # Remove location_id and set timestamp as index
        df.drop(columns=["location_id"], inplace=True, errors="ignore")
        df.set_index("timestamp", inplace=True)
        return df


class DemandSubset(Subset):
    """Demand subset of the HuggingFace dataset containing all demand data.

    Args:
        file_path: The path to the subset file.
    """

    def get_by_unique_id(self, unique_id: str):
        """Returns the demand data for the given unique id.

        Args:
            unique_id: The unique id of the meter.
        """
        query = """
            SELECT *
            FROM read_parquet(?)
            WHERE unique_id = ?
        """
        conn = duckdb.connect(":memory:")
        df = conn.execute(query, parameters=[str(self.file_path), unique_id]).fetch_df()
        if df.empty:
            msg = f"Unique id '{unique_id}' was not found."
            raise KeyError(msg)

        # Remove unique_id and set timestamp as index
        df.drop(columns=["unique_id"], inplace=True, errors="ignore")
        df.set_index("timestamp", inplace=True)
        return df


class DemandDataset:
    """DemandDataset class representing the HuggingFace dataset.

    This class is a collection of all subsets inside HuggingFace dataset.
    It provides an easy way to access the different subsets.

    Args:
        directory: The directory where the HuggingFace dataset is located.
                   This directory should contain all the subset files.
    """

    HUGGINGFACE_DATASET = "attila-balint-kul/electricity-demand"
    SUBSETS = ("demand", "metadata", "weather")

    def __init__(self, directory: Path | str) -> None:
        directory = Path(directory).resolve()
        if not directory.is_dir() or not directory.exists():
            msg = f"Please provide the existing directory where the '{self.HUGGINGFACE_DATASET}' dataset is located."
            raise ValueError(msg)
        self.directory = directory.resolve()

    def __repr__(self) -> str:
        return f"DemandDataset(directory={self.directory})"

    def _check_for_valid_subset(self, subset: str):
        if subset not in self.SUBSETS:
            msg = f"Please provide a valid subset. Available subsets: {self.SUBSETS}"
            raise ValueError(msg)

    @property
    def metadata_subset(self) -> MetadataSubset:
        """Returns the metadata subset."""
        return MetadataSubset(self._get_subset_path("metadata"))

    @property
    def weather_subset(self) -> WeatherSubset:
        """Returns the weather subset."""
        return WeatherSubset(self._get_subset_path("weather"))

    @property
    def demand_subset(self) -> DemandSubset:
        """Returns the demand subset."""
        return DemandSubset(self._get_subset_path("demand"))

    def get_subset(self, subset: str) -> Subset:
        """Returns the selected subset."""
        self._check_for_valid_subset(subset)
        if subset == "metadata":
            return self.metadata_subset
        elif subset == "weather":
            return self.weather_subset
        elif subset == "demand":
            return self.demand_subset
        msg = f"Please provide a valid subset. Available subsets: {self.SUBSETS}"
        raise ValueError(msg)

    def _get_subset_path(self, subset: str) -> Path:
        filepath = self.directory / f"{subset}.parquet"
        if not filepath.exists():
            msg = (
                f"There is no {subset} in the directory. "
                f"Make sure to download all subsets from the HuggingFace dataset: {self.HUGGINGFACE_DATASET}."
            )
            raise ValueError(msg)
        return self.directory / f"{subset}.parquet"

    def read_subset(self, subset: str) -> pd.DataFrame:
        """Reads the selected subset."""
        return self.get_subset(subset).read()

    def list_unique_ids(self) -> list[str]:
        return self.metadata_subset.list_unique_ids()

    def list_location_ids(self) -> list[str]:
        return self.weather_subset.list_location_ids()

    def get_data_by_unique_id(self, unique_id: str) -> tuple[pd.DataFrame | pd.DataFrame | Metadata]:
        metadata = self.metadata_subset.get_by_unique_id(unique_id)

        demand = self.demand_subset.get_by_unique_id(unique_id)
        weather = self.weather_subset.get_by_location_id(metadata.location_id)
        return demand, weather, metadata


class Dataset:
    """Dataset class representing a collection of data required for forecasting task.

    Args:
        target: The target variable.
        past_covariates: The past covariates.
        future_covariates: The future covariates.
        metadata: The metadata.
    """

    def __init__(
        self,
        target: pd.DataFrame,
        past_covariates: pd.DataFrame | None = None,
        future_covariates: pd.DataFrame | None = None,
        metadata: Metadata | None = None,
    ):
        self._target = self._check_target(target.copy())
        self._first_available_target_date: pd.Timestamp = self._target.index[0]
        self._last_available_target_date: pd.Timestamp = self._target.index[-1]

        self._past_covariates = (
            self._check_past_covariates(past_covariates.copy()) if past_covariates is not None else None
        )

        self._future_covariates = (
            self._check_external_forecasts(future_covariates.copy()) if future_covariates is not None else None
        )
        self.metadata = metadata

    @property
    def target_available_since(self) -> pd.Timestamp:
        """Returns the first available target date."""
        return self._first_available_target_date

    @property
    def target_available_until(self) -> pd.Timestamp:
        """Returns the last available target date."""
        return self._last_available_target_date

    @property
    def target_freq(self) -> str:
        """Returns the frequency of the target."""
        return self._target.index.inferred_freq

    @staticmethod
    def _check_target(y: pd.DataFrame) -> pd.DataFrame:
        if isinstance(y, pd.Series):
            logger.warning("Target is a Series, converting to DataFrame.")
            y = y.to_frame("y")
        if not isinstance(y.index, pd.DatetimeIndex):
            msg = f"Target dataframe must have DatetimeIndex, have {y.index.__class__.__name__}."
            raise ValueError(msg)
        y.rename_axis("timestamp", inplace=True)
        y.sort_index(inplace=True)
        return y

    def _check_past_covariates(self, X: pd.DataFrame) -> pd.DataFrame:  # noqa: N803
        if not isinstance(X.index, pd.DatetimeIndex):
            msg = f"Past covariates must have DatetimeIndex, have {X.index.__class__.__name__}."
            raise ValueError(msg)

        X.rename_axis("timestamp", inplace=True)
        X.sort_index(inplace=True)

        if X.index[0] > self._first_available_target_date:
            msg = "Covariates must be provided for the full target timeframe, covariates start after target values."
            raise ValueError(msg)

        if X.index[-1] < self._last_available_target_date:
            msg = "Covariates must be provided for the full target timeframe, covariates end before target values."
            raise ValueError(msg)
        return X

    def _check_external_forecasts(self, X: pd.DataFrame) -> pd.DataFrame:  # noqa: N803
        first_forecast_date = X.cutoff_date.min()
        last_forecast_date = X.cutoff_date.max()
        last_forecast_end_date = X[X.cutoff_date == last_forecast_date].timestamp.max()

        if first_forecast_date > self._first_available_target_date:
            msg = (
                "External forecasts must be provided for the full target timeframe, "
                "forecasts start after target values."
            )
            raise ValueError(msg)

        if last_forecast_end_date < self._last_available_target_date:
            msg = (
                "External forecasts must be provided for the full target timeframe, "
                "forecasts end before target values."
            )
            raise ValueError(msg)

        return X

    def _check_cutoff_in_rage(self, cutoff_date: pd.Timestamp):
        if cutoff_date < self._first_available_target_date:
            msg = f"Cutoff date is before the start date: {cutoff_date} < {self._first_available_target_date}."
            raise IndexError(msg)

        if cutoff_date > self._last_available_target_date:
            msg = f"Cutoff date is after the end date: {cutoff_date} > {self._last_available_target_date}."
            raise IndexError(msg)

    def get_history(self, cutoff_date: pd.Timestamp) -> pd.DataFrame:
        """Returns the history of the target variable up to the cutoff date.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The history of the target variable up to the cutoff date.
        """
        self._check_cutoff_in_rage(cutoff_date)
        return self._target[self._target.index <= cutoff_date]

    def get_past_covariates(self, cutoff_date: pd.Timestamp) -> pd.DataFrame | None:
        """Returns the past covariates for the cutoff date.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The past covariates up until the cutoff date.
        """
        if self._past_covariates is None:
            return None

        self._check_cutoff_in_rage(cutoff_date)
        return self._past_covariates[self._past_covariates.index <= cutoff_date]

    def get_future_covariates(self, cutoff_date: pd.Timestamp) -> pd.DataFrame | None:
        """Returns the future covariates for the cutoff date.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The last external forecasts made before the cutoff date.
        """
        if self._future_covariates is None:
            return None

        self._check_cutoff_in_rage(cutoff_date)
        last_past_cutoff_date = self._future_covariates.cutoff_date[
            self._future_covariates.cutoff_date <= cutoff_date
        ].max()

        future_covariates = self._future_covariates[
            (self._future_covariates.cutoff_date == last_past_cutoff_date)
            & (self._future_covariates.timestamp > cutoff_date)
        ]
        future_covariates.set_index("timestamp", inplace=True)
        return future_covariates



from http import HTTPStatus

import numpy as np
import pandas as pd
import pytest
from starlette.testclient import TestClient

from enfobench.evaluation.client import to_buffer
from enfobench.evaluation.server import server_factory


@pytest.fixture(scope="function")
def forecast_client(model):
    app = server_factory(model)
    client = TestClient(app)
    return client


def test_info_endpoint(forecast_client):
    response = forecast_client.get("/info")

    assert response.status_code == HTTPStatus.OK
    assert response.json() == {
        "name": "TestModel",
        "authors": [
            {
                "name": "Author 1",
                "email": "author-1@institution.org",
            }
        ],
        "type": "point",
        "params": {
            "param1": 1,
        },
    }


def test_environment_endpoint(forecast_client):
    response = forecast_client.get("/environment")

    assert response.status_code == HTTPStatus.OK
    assert "packages" in response.json()
    for package_name, package_version in response.json()["packages"].items():
        assert isinstance(package_name, str)
        assert isinstance(package_version, str)


def test_forecast_endpoint(forecast_client):
    horizon = 24
    target_index = pd.date_range(start="2020-01-01", end="2021-01-01", freq="1H")
    history_df = pd.DataFrame(
        index=target_index,
        data={
            "y": np.random.normal(size=len(target_index)),
        },
    )

    response = forecast_client.post(
        "/forecast",
        params={
            "horizon": horizon,
        },
        files={
            "history": to_buffer(history_df),
        },
    )
    assert response.status_code == HTTPStatus.OK
    assert "forecast" in response.json()
    forecast = response.json()["forecast"]
    assert isinstance(forecast, list) and len(forecast) == horizon
    for forecast_item in forecast:
        assert "timestamp" in forecast_item and pd.Timestamp(forecast_item["timestamp"])
        assert "yhat" in forecast_item and isinstance(
            forecast_item["yhat"],
            float | int,
        )



import io
import sys
from typing import Annotated, Any

import pandas as pd
import pkg_resources
from fastapi import FastAPI, File, Query
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from starlette.responses import RedirectResponse

from enfobench.evaluation.model import ForecasterType, Model


class AuthorInfo(BaseModel):
    """Author information.

    Attributes:
        name: Name of the author.
        email: Email of the author.
    """

    name: str
    email: str | None = None


class ModelInfo(BaseModel):
    """Model information.

    Attributes:
        name: Name of the model.
        authors: List of authors.
        type: Type of the model.
        params: Parameters of the model.
    """

    name: str
    authors: list[AuthorInfo]
    type: ForecasterType  # noqa: A003
    params: dict[str, Any]


class EnvironmentInfo(BaseModel):
    python: str
    packages: dict[str, str]


def server_factory(model: Model) -> FastAPI:
    app = FastAPI()
    environment = EnvironmentInfo(
        python=sys.version,
        packages={package.key: package.version for package in pkg_resources.working_set},
    )

    @app.get("/", include_in_schema=False)
    async def index():
        return RedirectResponse(url="/docs")

    @app.get("/info", response_model=ModelInfo)
    async def model_info():
        """Return model information."""
        return model.info()

    @app.get("/environment", response_model=EnvironmentInfo)
    async def environment_info():
        """Return information of installed packages and their versions."""
        return environment

    @app.post("/forecast")
    async def forecast(
        horizon: int,
        history: Annotated[bytes, File()],
        past_covariates: Annotated[bytes | None, File()] = None,
        future_covariates: Annotated[bytes | None, File()] = None,
        level: list[int] | None = Query(None),  # noqa: B008
    ):
        history_df = pd.read_parquet(io.BytesIO(history))
        past_covariates_df = pd.read_parquet(io.BytesIO(past_covariates)) if past_covariates is not None else None
        future_covariates_df = pd.read_parquet(io.BytesIO(future_covariates)) if future_covariates is not None else None

        forecast_df = model.forecast(
            horizon=horizon,
            history=history_df,
            past_covariates=past_covariates_df,
            future_covariates=future_covariates_df,
            level=level,
        )
        forecast_df.fillna(0, inplace=True)
        forecast_df.rename_axis("timestamp", inplace=True)
        forecast_df.reset_index(inplace=True)

        response = {
            "forecast": jsonable_encoder(forecast_df.to_dict(orient="records")),
        }
        return JSONResponse(
            content=response,
            status_code=200,
        )

    return app



import warnings
from collections.abc import Callable
from typing import Any

import numpy as np
import pandas as pd
from tqdm import tqdm

from enfobench.dataset import Dataset
from enfobench.evaluation.client import ForecastClient
from enfobench.evaluation.model import Model
from enfobench.evaluation.utils import generate_cutoff_dates, steps_in_horizon


def _compute_metric(forecast: pd.DataFrame, metric: Callable) -> float:
    """Compute a single metric value.

    Args:
        forecast: Forecast to evaluate.
        metric: Metric to evaluate.

    Returns:
        Metric value.
    """
    metric_value = metric(forecast.y, forecast.yhat)
    return metric_value


def _compute_metrics(forecast: pd.DataFrame, metrics: dict[str, Callable]) -> dict[str, float]:
    """Compute multiple metric values.

    Args:
        forecast: Forecast to evaluate.
        metrics: Metric to evaluate.

    Returns:
        Metrics dictionary with metric names as keys and metric values as values.
    """
    metric_values = {metric_name: _compute_metric(forecast, metric) for metric_name, metric in metrics.items()}
    return metric_values


def _evaluate_group(forecasts: pd.DataFrame, metrics: dict[str, Callable], index: Any) -> pd.DataFrame:
    clean_df = forecasts.dropna(subset=["y"])
    if clean_df.empty:
        ratio = 0.0
        metrics = {metric_name: np.nan for metric_name in metrics}
    else:
        ratio = len(clean_df) / len(forecasts)
        metrics = _compute_metrics(clean_df, metrics)
    df = pd.DataFrame({**metrics, "weight": ratio}, index=[index])
    return df


def evaluate_metrics(
    forecasts: pd.DataFrame,
    metrics: dict[str, Callable],
    *,
    groupby: str | None = None,
) -> pd.DataFrame:
    """Evaluate multiple metrics on forecasts.

    Args:
        forecasts: Forecasts to evaluate.
        metrics: Metric to evaluate.
        groupby: Column to group forecasts by. (Optional, if not provided, forecasts will not be grouped.)

    Returns:
        Metric values for each cutoff with their weight.
    """
    if groupby is None:
        return _evaluate_group(forecasts, metrics, 0)

    if groupby not in forecasts.columns:
        msg = f"Groupby column {groupby} not found in forecasts."
        raise ValueError(msg)

    metrics_df = pd.concat(
        [_evaluate_group(group_df, metrics, value) for value, group_df in tqdm(forecasts.groupby(groupby))]
    )

    metrics_df.rename_axis(groupby, inplace=True)
    metrics_df.reset_index(inplace=True)
    return metrics_df


def cross_validate(
    model: Model | ForecastClient,
    dataset: Dataset,
    *,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    horizon: pd.Timedelta,
    step: pd.Timedelta,
    level: list[int] | None = None,
) -> pd.DataFrame:
    """Cross-validate a model.

    Args:
        model: Model to cross-validate.
        dataset: Dataset to cross-validate on.
        start_date: Start date of the time series.
        end_date: End date of the time series.
        horizon: Forecast horizon.
        step: Step size between cutoff dates.
        level: Prediction intervals to compute. (Optional, if not provided, simple point forecasts will be computed.)
    """
    if start_date <= dataset.target_available_since:
        msg = f"Start date must be after the start of the dataset: {start_date} <= {dataset.target_available_since}."
        raise ValueError(msg)

    initial_training_data = start_date - dataset.target_available_since
    if initial_training_data < pd.Timedelta("7 days"):
        warnings.warn("Initial training data is less than 7 days.", stacklevel=2)

    if end_date > dataset.target_available_until:
        msg = f"End date must be before the end of the dataset: {end_date} > {dataset.target_available_until}."
        raise ValueError(msg)

    cutoff_dates = generate_cutoff_dates(start_date, end_date, horizon, step)
    horizon_length = steps_in_horizon(horizon, dataset.target_freq)

    forecasts = []
    for cutoff_date in tqdm(cutoff_dates):
        history = dataset.get_history(cutoff_date)
        past_covariates = dataset.get_past_covariates(cutoff_date)
        future_covariates = dataset.get_future_covariates(cutoff_date)

        forecast = model.forecast(
            horizon_length,
            history=history,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            level=level,
        )
        if not isinstance(forecast, pd.DataFrame) or not isinstance(forecast.index, pd.DatetimeIndex):
            msg = (
                f"Forecast must be a DataFrame with a DatetimeIndex, "
                f"got {type(forecast)} with index {type(forecast.index)}."
            )
            raise ValueError(msg)

        forecast_contains_nans = forecast.isna().any(axis=None)
        if forecast_contains_nans:
            msg = "Forecast contains NaNs, make sure to fill in missing values."
            raise ValueError(msg)

        if len(forecast) != horizon_length:
            msg = f"Forecast does not match the requested horizon length {horizon_length}, got {len(forecast)}."
            raise ValueError(msg)

        forecast.rename_axis("timestamp", inplace=True)
        forecast.reset_index(inplace=True)
        forecast["cutoff_date"] = pd.to_datetime(cutoff_date, unit="ns")
        forecast.set_index(["cutoff_date", "timestamp"], inplace=True)
        forecasts.append(forecast)

    crossval_df = pd.concat(forecasts).reset_index()

    # Merge the forecast with the target
    crossval_df = crossval_df.merge(dataset._target, left_on="timestamp", right_index=True)
    return crossval_df



import numpy as np
import pytest

from enfobench.evaluation.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_bias_error,
    mean_squared_error,
    root_mean_squared_error,
)

all_metrics = [
    mean_absolute_error,
    mean_bias_error,
    root_mean_squared_error,
    mean_squared_error,
    mean_absolute_percentage_error,
]


@pytest.mark.parametrize("metric", all_metrics)
def test_metric_raises_with_nans(metric):
    with pytest.raises(ValueError):
        assert metric(np.array([1, 2, 3]), np.array([np.nan, 2, 3])) == 0

    with pytest.raises(ValueError):
        assert metric(np.array([1, np.nan, 3]), np.array([1, 2, 3])) == 0


@pytest.mark.parametrize("metric", all_metrics)
def test_metric_raises_with_unequal_length(metric):
    with pytest.raises(ValueError):
        assert metric(np.array([1, 2, 3]), np.array([])) == 0


@pytest.mark.parametrize("metric", all_metrics)
def test_metric_raises_with_empty_array(metric):
    with pytest.raises(ValueError):
        assert metric(np.array([1, 2, 3]), np.array([])) == 0

    with pytest.raises(ValueError):
        assert metric(np.array([]), np.array([1, 2, 3])) == 0


@pytest.mark.parametrize(
    "a,b,value",
    [
        (np.array([1, 2, 3]), np.array([1, 2, 3]), 0.0),
        (np.array([1, 2, 3]), np.array([2, 3, 4]), 1.0),
        (np.array([1, 2, 3]), np.array([0, 1, 2]), 1.0),
    ],
)
def test_mean_absolute_error(a, b, value):
    assert mean_absolute_error(a, b) == value


@pytest.mark.parametrize(
    "a,b,value",
    [
        (np.array([1, 2, 3]), np.array([1, 2, 3]), 0.0),
        (np.array([1, 2, 3]), np.array([2, 3, 4]), 1.0),
        (np.array([1, 2, 3]), np.array([0, 1, 2]), -1.0),
    ],
)
def test_mean_bias_error(a, b, value):
    assert mean_bias_error(a, b) == value



from random import randrange

import pandas as pd
import pytest

from enfobench.dataset import Dataset


def random_date(start: pd.Timestamp, end: pd.Timestamp, resolution: int = 1) -> pd.Timestamp:
    """
    This function will return a random datetime between two datetime
    objects.
    """
    delta = end - start
    int_delta = int(delta.total_seconds())
    random_second = randrange(0, int_delta, resolution)  # noqa: S311
    return start + pd.Timedelta(seconds=random_second)


@pytest.mark.parametrize("resolution", [1, 60, 900, 3600])
def test_get_history(target, resolution):
    ds = Dataset(target=target)
    cutoff_date = random_date(ds._first_available_target_date, ds._last_available_target_date, resolution=resolution)

    history = ds.get_history(cutoff_date)

    assert history.index.name == "timestamp"
    assert isinstance(history.index, pd.DatetimeIndex)
    assert "y" in history.columns
    assert len(history.columns) == 1
    assert (history.index <= cutoff_date).all()


@pytest.mark.parametrize("resolution", [1, 60, 900, 3600])
def test_get_past_covariates(target, covariates, resolution):
    ds = Dataset(target=target, past_covariates=covariates)
    cutoff_date = random_date(ds._first_available_target_date, ds._last_available_target_date, resolution=resolution)

    past_cov = ds.get_past_covariates(cutoff_date)

    assert past_cov.index.name == "timestamp"
    assert isinstance(past_cov.index, pd.DatetimeIndex)
    for col in covariates.columns:
        assert col in past_cov.columns
    assert (past_cov.index <= cutoff_date).all()


@pytest.mark.parametrize("resolution", [1, 60, 900, 3600])
def test_get_future_covariates(target, covariates, external_forecasts, resolution):
    ds = Dataset(target=target, past_covariates=covariates, future_covariates=external_forecasts)
    cutoff_date = random_date(ds._first_available_target_date, ds._last_available_target_date, resolution=resolution)

    future_cov = ds.get_future_covariates(cutoff_date)

    assert future_cov.index.name == "timestamp"
    assert isinstance(future_cov.index, pd.DatetimeIndex)
    for col in external_forecasts.columns:
        if col not in ["timestamp", "cutoff_date"]:
            assert col in future_cov.columns
    assert (future_cov.cutoff_date <= cutoff_date).all()
    assert (future_cov.index > cutoff_date).all()



from enfobench.dataset._dataset import Dataset, DemandDataset, DemandSubset, MetadataSubset, WeatherSubset

__all__ = [
    "Dataset",
    "DemandDataset",
    "DemandSubset",
    "MetadataSubset",
    "WeatherSubset",
]
